@@ -1,6 +1,6 @@
-use crate::ext::{AsOptExpr, MapWithMut, PatOrExprExt};
+use crate::ext::{AsOptExpr, PatOrExprExt};
 use fxhash::FxHashMap;
-use swc_common::{comments::Comments, Span, Spanned};
+use swc_common::{comments::Comments, util::take::Take, Span, Spanned};
 use swc_ecma_ast::*;
 use swc_ecma_visit::{as_folder, noop_visit_mut_type, Fold, VisitMut, VisitMutWith};
 
